from __future__ import annotations

import subprocess
import sys
import zipfile
from contextlib import contextmanager
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING

import toml

import hatch_build_scripts

if TYPE_CHECKING:
    from collections.abc import Iterator
    from collections.abc import Sequence

    from hatch_build_scripts.plugin import OneScriptConfig

ROOT_DIR = Path(__file__).parent.parent


def create_project(path: Path | str, scripts: Sequence[OneScriptConfig]) -> FakeProject:
    path = Path(path)

    full_config = {
        "project": {
            "name": "test-project",
            "version": "0.0.0",
            "description": "A test project",
        },
        "build-system": {
            "requires": ["hatchling", f"hatch-build-scripts @ {ROOT_DIR.as_uri()}"],
            "build-backend": "hatchling.build",
        },
        "tool": {
            "hatch": {
                "build": {
                    "hooks": {
                        "build-scripts": {
                            "scripts": list(map(asdict, scripts)),
                        }
                    }
                }
            }
        },
    }

    (path / "pyproject.toml").write_text(toml.dumps(full_config), encoding="utf-8")

    return FakeProject(path)


class FakeProject:
    def __init__(self, path: Path) -> None:
        self.path = path

    def build(self) -> None:
        subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "cache",
                "remove",
                hatch_build_scripts.__name__,
            ],
            cwd=self.path,
            check=True,
        )
        subprocess.run(
            [sys.executable, "-m", "hatch", "build"],
            cwd=self.path,
            check=True,
        )

    @contextmanager
    def dist(self) -> Iterator[zipfile.ZipFile]:
        files = list((self.path / "dist").glob("*.whl"))
        assert len(files) == 1
        with zipfile.ZipFile(str(files[0])) as whl:
            yield whl



from hatch_build_scripts.plugin import OneScriptConfig
from tests.utils import create_project


def test_plugin(tmpdir):
    tmp_lib_dir = tmpdir / "lib"
    tmp_lib_dir.mkdir()

    (tmp_lib_dir / "some-dir").mkdir()
    (tmp_lib_dir / "another-dir").mkdir()
    (tmp_lib_dir / "yet-another-dir").mkdir()

    some_dir_out = tmp_lib_dir / "some-dir-out"
    some_dir_out.mkdir()
    # we expect that this file will not be cleaned
    (some_dir_out / "module.py").write_text('print("hello")', "utf-8")
    # we expect that this file will be cleaned
    (some_dir_out / "f3.txt").write_text("this should be cleaned", "utf-8")

    another_dir_out = tmp_lib_dir / "another-dir-out"
    another_dir_out.mkdir()
    # we expect that this file will be cleaned
    (another_dir_out / "module.py").write_text('print("hello")', "utf-8")

    proj = create_project(
        tmp_lib_dir,
        [
            OneScriptConfig(
                out_dir="fake",
                commands=["echo 'hello world' > fake.txt"],
                artifacts=["fake.txt"],
            ),
            OneScriptConfig(
                out_dir="some-dir-out",
                work_dir="some-dir",
                commands=[
                    "echo 'hello world' > f1.txt",
                    "echo 'hello world' > f2.txt",
                ],
                # this will not clean the data.json file
                artifacts=["*.txt"],
            ),
            OneScriptConfig(
                out_dir="another-dir-out",
                work_dir="another-dir",
                commands=[
                    "echo 'hello world' > f1.txt",
                    "echo 'hello world' > f2.txt",
                ],
                artifacts=["*.txt"],
                clean_out_dir=True,
            ),
            OneScriptConfig(
                out_dir="yet-another-dir-out",
                work_dir="yet-another-dir",
                commands=[
                    "echo 'hello world' > f1.txt",
                    "echo 'hello world' > f2.txt",
                ],
                artifacts=["*.txt"],
                clean_out_dir=True,
                clean_artifacts_after_build=True,
            ),
        ],
    )

    proj.build()

    extract_dir = tmpdir / "extract"
    extract_dir.mkdir()

    with proj.dist() as dist:
        dist.extractall(extract_dir)

        assert (extract_dir / "fake" / "fake.txt").exists()
        assert (extract_dir / "some-dir-out" / "module.py").exists()

        assert not (extract_dir / "some-dir-out" / "f3.txt").exists()
        assert not (extract_dir / "another-dir-out" / "module.py").exists()

        # we expect that this file still exists in the source
        assert (tmp_lib_dir / "fake" / "fake.txt").exists()
        # we expect that this file was cleaned in the source but exists in wheel
        assert (extract_dir / "yet-another-dir-out" / "f1.txt").exists()
        assert not (tmp_lib_dir / "yet-another-dir-out" / "f1.txt").exists()



# ruff: noqa: S607,S603,S404,FBT001,D401,D103
from __future__ import annotations

import os
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Literal

import click

IN_CI = bool(os.getenv("GITHUB_ACTIONS"))


@click.group()
def main():
    """A collection of dev utilities."""


@main.command("test")
@click.argument("args", nargs=-1)
def test(args: list[str]):
    """Run the test suite."""
    run(["pytest", "-v", *args])


@main.command("cov")
@click.option("--no-test", is_flag=True, help="Skip running tests with coverage")
@click.option("--old-coverage-xml", default=None, type=str, help="Path to target coverage.xml.")
def cov(no_test: bool, old_coverage_xml: str | None):
    """Run the test suite with coverage."""
    if not no_test:
        try:
            run(["coverage", "run", "-m", "pytest", "-v"])
        finally:
            run(["coverage", "combine"], check=False)
            run(["coverage", "report"])
            run(["coverage", "xml"])
    if old_coverage_xml is not None:
        if Path(old_coverage_xml).exists():
            run(
                [
                    "pycobertura",
                    "diff",
                    old_coverage_xml,
                    "coverage.xml",
                    "--source1",
                    ".",
                    "--source2",
                    ".",
                ]
            )
        else:
            msg = f"Target coverage file {old_coverage_xml} does not exist"
            raise click.ClickException(msg)
    elif not IN_CI:
        run(["diff-cover", "coverage.xml", "--config-file", "pyproject.toml"])


@main.command("lint")
@click.option("--check", is_flag=True, help="Check for linting issues without fixing.")
@click.option("--no-py-style", is_flag=True, help="Skip style check Python files.")
@click.option("--no-py-types", is_flag=True, help="Skip type check Python files.")
@click.option("--no-uv-locked", is_flag=True, help="Skip check that the UV lock file is synced")
@click.option("--no-yml-style", is_flag=True, help="Skip style check YAML files.")
def lint(
    check: bool,
    no_py_style: bool,
    no_py_types: bool,
    no_uv_locked: bool,
    no_yml_style: bool,
):
    """Linting commands."""
    if not no_uv_locked:
        run(["uv", "lock", "--locked"])
    if not no_py_style:
        if check:
            run(["ruff", "format", "--check", "--diff"])
            run(["ruff", "check"])
        else:
            run(["ruff", "format"])
            run(["ruff", "check", "--fix"])
    if not no_yml_style:
        if check:
            run(["yamlfix", "--check", ".github"])
        else:
            run(["yamlfix", ".github"])
    if not no_py_types:
        run(["pyright"])


if TYPE_CHECKING:
    from collections.abc import Sequence

    run = subprocess.run
else:

    def run(*args, **kwargs):
        cmd, *args = args
        cmd = tuple(map(str, cmd))
        kwargs.setdefault("check", True)
        click.echo(click.style(" ".join(cmd), bold=True))
        try:
            return subprocess.run(cmd, *args, **kwargs)
        except subprocess.CalledProcessError as e:
            raise click.ClickException(e) from None
        except FileNotFoundError as e:
            msg = f"File not found {e}"
            raise click.ClickException(msg) from None


def report(
    kind: Literal["notice", "warning", "error"],
    /,
    *,
    title: str = "",
    message: str = "",
    file: str | None = None,
    line: int | None = None,
    end_line: int | None = None,
    col: int | None = None,
    end_col: int | None = None,
):
    if not IN_CI:
        file_parts = []
        if file:
            file_parts.append(f"{file}")
            if line:
                file_parts.append(f":{line}")
                if end_line:
                    file_parts.append(f"-{end_line}")
            if col:
                file_parts.append(f":{col}")
                if end_col:
                    file_parts.append(f"-{end_col}")
        file_info = "".join(file_parts)
        click.echo(" - ".join(filter(None, [kind.upper(), file_info, title, message])))
    else:
        file_parts = []
        if title or message:
            file_parts.append(f"{title}::{message}")
        if file:
            file_parts.append(f"file={file}")
            if line:
                file_parts.append(f"line={line}")
                if end_line:
                    file_parts.append(f"endLine={end_line}")
            if col:
                file_parts.append(f"col={col}")
                if end_col:
                    file_parts.append(f"endCol={end_col}")
        click.echo(f"::{kind} {','.join(file_parts)}")


def doc_cmd(cmd: Sequence[str], *, no_pad: bool = False):
    run(
        list(
            filter(
                None,
                [
                    "doccmd",
                    "-v",
                    "--language=python",
                    "--no-pad-file" if no_pad else "",
                    "--command",
                    " ".join(cmd),
                    "docs",
                ],
            )
        )
    )


if __name__ == "__main__":
    main()
